<?php

namespace Phine\Phar\Signature\Algorithm;

use Phine\Phar\File\Reader;

/**
 * Provides the basis for a `hash()` supported signature algorithm.
 *
 * @author Kevin Herrera <kevin@herrera.io>
 *
 * @api
 */
abstract class AbstractHashAlgorithm implements AlgorithmInterface
{
    /**
     * {@inheritDoc}
     */
    public function readSignature(Reader $reader)
    {
        $size = $this->getSize();

        $reader->seek(-8 - $size, SEEK_END);

        $hash = unpack('H*', $reader->read($size));
        $hash = strtoupper($hash[1]);

        return array(
            'hash' => $hash,
            'hash_type' => $this->getName()
        );
    }

    /**
     * {@inheritDoc}
     */
    public function verifySignature(Reader $reader)
    {
        $expected = $this->readSignature($reader);

        $context = hash_init($this->getAlgorithm());

        $reader->seek(0);

        hash_update_stream(
            $context,
            $reader->getHandle(),
            $reader->getSize() - $this->getSize() - 8
        );

        $actual = strtoupper(hash_final($context));

        return ($expected['hash'] === $actual);
    }

    /**
     * Returns the algorithm name as recognized by `hash()`.
     *
     * This method will return the name of the hash algorithm that will be
     * used with the `hash()` function. Note that you may want to check the
     * returned name against `hash_algos()` to see if the algorithm is in
     * fact supported.
     *
     *     echo $this->getAlgorithm(); // sha1
     *
     * @return string The name of the algorithm.
     *
     * @api
     */
    abstract protected function getAlgorithm();

    /**
     * Returns the name of the algorithm as returned by `Phar::getHash()`.
     *
     * This method will return the name of the hash algorithm as it would
     * have been returned by the `Phar::getSignature()` method. The case
     * and formatting may be different from that of ones returned by the
     * `hash_algos()` method.
     *
     *     echo $this->getName(); // SHA-1
     *
     * @return string The name of the algorithm.
     *
     * @api
     */
    abstract protected function getName();

    /**
     * Returns the byte size of the hash generated by the algorithm.
     *
     * This method will return the number of bytes that will need to be read
     * from the archive file in order to extract the whole signature hash (not
     * including the type flag and "GBMB" magic flag).
     *
     *     echo $this->getSize(); // 20
     *
     * @return integer The size of the hash.
     *
     * @api
     */
    abstract protected function getSize();
}
